/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include <tapioca/base/tapioca_enums.h>

#include "tapiocaui_buddy_list.h"
#include "tapiocaui_cell_renderer_pixbuf.h"
#include "tapiocaui_db.h"
#include "tapiocaui_marshal.h"
#include "tapiocaui_utils.h"

#include <string.h>

/*
static const gchar *presence_status_desc[] = {
    "Offline",
    "Available",
    "Away From Computer",
    "Extended Away",
    "Busy",
    NULL
};
*/
static const gushort presence_priority[] = { 4, 0, 2, 3, 1 };

enum {
    BUDDY_COLUMN,
    STATUS_COLUMN,
    STATUS_ICON_COLUMN, 
    NAME_COLUMN,
    CALL_COLUMN,
    CALL_SENSITIVE_COLUMN,
    LAST_COLUMN
};

enum
{
    PROP_0,
    PROP_DB
};

enum
{
    BUDDY_ACTIVATED,
    BUDDY_SELECTION_CHANGED,
    BUDDY_REMOVE_REQUESTED,
    LAST_SIGNAL
};

struct _TapiocaUIBuddyListPrivate
{
    TapiocaUIDB *db;
    GtkListStore *buddy_store;
    GtkTreeViewColumn *column_name;
    GtkTreeViewColumn *column_call;
    gint old_width;

    GdkPixbuf *img_call_active;
    GdkPixbuf *img_status[TAPIOCA_PRESENCE_STATUS_LAST + 1];
    GtkWidget *menu;
    GtkWidget *chat_menuitem;
    GtkWidget *call_menuitem;
};

static void       tapiocaui_buddy_list_class_init               (TapiocaUIBuddyListClass *k);
static void       tapiocaui_buddy_list_init                     (TapiocaUIBuddyList *self);
static GObject   *tapiocaui_buddy_list_constructor              (GType type,
                                                                 guint n_construct_properties,
                                                                 GObjectConstructParam *construct_properties);
static void       tapiocaui_buddy_list_finalize                 (GObject *object);

static gboolean   tapiocaui_buddy_list_search_func              (GtkTreeModel *model,
                                                                 gint column,
                                                                 const gchar *key,
                                                                 GtkTreeIter *iter,
                                                                 gpointer search_data);
static gchar     *tapiocaui_buddy_list_get_description          (TapiocaUIBuddyData *buddy);
static void       tapiocaui_buddy_list_iter_selected_cb         (TapiocaUIList *self,
                                                                 GtkTreeIter *iter,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_iter_unselected_cb       (TapiocaUIList *self,
                                                                 GtkTreeIter *iter,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_iter_activated_cb        (TapiocaUIList *self,
                                                                 GtkTreeIter *iter,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_db_changed_cb            (TapiocaUIDB *db,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_db_buddy_added_cb        (TapiocaUIDB *db,
                                                                 gpointer buddy,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_db_buddy_removed_cb      (TapiocaUIDB *db,
                                                                 gpointer buddy,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_img_call_activate_cb     (TapiocaUICellRendererPixbuf *cell,
                                                                 const gchar *path,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_resize_cb                (GtkWidget *widget,
                                                                 GtkAllocation *allocation,
                                                                 gpointer user_data);
static void       tapiocaui_buddy_list_sort_buddy               (TapiocaUIBuddyList *self,
                                                                 TapiocaUIBuddyData *buddy,
                                                                 guint priority,
                                                                 GtkTreeIter *iter);
static gboolean   tapiocaui_buddy_list_get_buddy_iter           (TapiocaUIBuddyList *self,
                                                                 TapiocaUIBuddyData *buddy,
                                                                 GtkTreeIter *iter);
static void       tapiocaui_buddy_list_insert_sorted            (TapiocaUIBuddyList *self,
                                                                 TapiocaUIBuddyData *buddy,
                                                                 guint priority,
                                                                 GtkTreeIter *iter);

/* popup menu callbacks */
static void       tapiocaui_buddy_list_create_popup_menu         (TapiocaUIBuddyList *self);
static void       tapiocaui_buddy_list_menu_item_open_chat_activated_cb
                                                                 (GtkWidget *menuitem,
                                                                  gpointer user_data);
static void       tapiocaui_buddy_list_menu_item_call_activated_cb
                                                                 (GtkWidget *menuitem,
                                                                  gpointer user_data);
static void       tapiocaui_buddy_list_menu_item_remove_activated_cb
                                                                 (GtkWidget *menuitem,
                                                                  gpointer user_data);
static gboolean   tapiocaui_buddy_list_button_pressed_cb         (GtkWidget *treeview,
                                                                  GdkEventButton *event,
                                                                  gpointer user_data);

static GObjectClass *parent_class = NULL;
static guint tapiocaui_buddy_list_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (TapiocaUIBuddyList, tapiocaui_buddy_list, TAPIOCAUI_TYPE_LIST)

static void
tapiocaui_buddy_list_class_init (TapiocaUIBuddyListClass *k)
{
    GObjectClass *object_class = G_OBJECT_CLASS (k);
    parent_class = g_type_class_peek_parent (k);
    object_class->constructor = tapiocaui_buddy_list_constructor;
    object_class->finalize = tapiocaui_buddy_list_finalize;

    tapiocaui_buddy_list_signals[BUDDY_ACTIVATED] =
        g_signal_new ("buddy-activated",
                      G_OBJECT_CLASS_TYPE (object_class),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIBuddyListClass, buddy_activated),
                      NULL, NULL,
                      tapiocaui_marshal_VOID__POINTER_BOOLEAN,
                      G_TYPE_NONE,
                      2,
                      G_TYPE_POINTER,
                      G_TYPE_BOOLEAN);
    tapiocaui_buddy_list_signals[BUDDY_SELECTION_CHANGED] =
        g_signal_new ("buddy-selection-changed",
                      G_OBJECT_CLASS_TYPE (object_class),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIBuddyListClass, buddy_selection_changed),
                      NULL, NULL,
                      tapiocaui_marshal_VOID__POINTER_BOOLEAN,
                      G_TYPE_NONE,
                      2,
                      G_TYPE_POINTER,
                      G_TYPE_BOOLEAN);
    tapiocaui_buddy_list_signals[BUDDY_REMOVE_REQUESTED] =
        g_signal_new ("buddy-remove-requested",
                      G_OBJECT_CLASS_TYPE (object_class),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (TapiocaUIBuddyListClass, buddy_remove_requested),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_POINTER);
}

static void
tapiocaui_buddy_list_init (TapiocaUIBuddyList *self)
{
    self->priv = g_new0 (TapiocaUIBuddyListPrivate, 1);
}

static GObject *
tapiocaui_buddy_list_constructor (GType type,
                                  guint n_construct_properties,
                                  GObjectConstructParam *construct_properties)
{
    TapiocaUIBuddyList *self;
    TapiocaUIBuddyListClass *klass;
    GObjectClass *parent_class;
    GtkTreeViewColumn *col;
    GtkCellRenderer *render;

    klass = TAPIOCAUI_BUDDY_LIST_CLASS (g_type_class_peek (TAPIOCAUI_TYPE_BUDDY_LIST));

    parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));

    self = TAPIOCAUI_BUDDY_LIST (parent_class->constructor (type, n_construct_properties,
                                                            construct_properties));

    self->priv->buddy_store = gtk_list_store_new (LAST_COLUMN,
                                                  G_TYPE_POINTER,
                                                  G_TYPE_INT, GDK_TYPE_PIXBUF,
                                                  G_TYPE_STRING, GDK_TYPE_PIXBUF,
                                                  G_TYPE_BOOLEAN);
    tapiocaui_list_set_model (TAPIOCAUI_LIST (self),
                              GTK_TREE_MODEL (self->priv->buddy_store));

#ifdef WITH_HILDON
    gtk_tree_view_set_rules_hint (tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self)), TRUE);

    self->priv->img_call_active = gdk_pixbuf_new_from_file (
            tapiocaui_file ("call.png"), NULL);

    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_AVAILABLE] = gdk_pixbuf_new_from_file (
            tapiocaui_file ("available.png"), NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_AWAY] = gdk_pixbuf_new_from_file (
            tapiocaui_file ("away.png"), NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_XA] = gdk_pixbuf_new_from_file (
            tapiocaui_file ("away.png"), NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_DND] = gdk_pixbuf_new_from_file (
            tapiocaui_file ("busy.png"), NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_OFFLINE] = gdk_pixbuf_new_from_file (
            tapiocaui_file ("offline.png"), NULL);

#else
    self->priv->img_call_active = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("call.png"), 21, 24, NULL);

    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_AVAILABLE] = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("available.png"), 12, 12, NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_AWAY] = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("away.png"), 12, 12, NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_XA] = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("away.png"), 12, 12, NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_DND] = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("busy.png"), 12, 12, NULL);
    self->priv->img_status[TAPIOCA_PRESENCE_STATUS_OFFLINE] = gdk_pixbuf_new_from_file_at_size (
            tapiocaui_file ("offline.png"), 12, 12, NULL);
#endif

    // BUDDY-STATUS-COLUMN
    col = gtk_tree_view_column_new ();

    render = gtk_cell_renderer_pixbuf_new ();
    gtk_tree_view_column_pack_start (col, render, FALSE);
    gtk_tree_view_column_set_attributes (col, render,
                                         "pixbuf", STATUS_ICON_COLUMN,
                                         NULL);


    // BUDDY-NAME-COLUMN
    render = gtk_cell_renderer_text_new ();
    gtk_tree_view_column_pack_start (col, render, TRUE);
    gtk_tree_view_column_set_attributes (col, render,
                                         "markup", NAME_COLUMN,
                                         NULL);
    g_object_set (render,
                  "ellipsize", PANGO_ELLIPSIZE_END,
                  NULL);
    gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
    tapiocaui_list_append_collumn (TAPIOCAUI_LIST (self), col);
    self->priv->column_name = col;

    // BUDDY-CALL-COLUMN
    col = gtk_tree_view_column_new ();
    render = tapiocaui_cell_renderer_pixbuf_new ();
    gtk_tree_view_column_pack_start (col, render, FALSE);
    g_object_set (render,
                  "mode", GTK_CELL_RENDERER_MODE_ACTIVATABLE,
                  "xalign", 0.5, NULL);
    gtk_tree_view_column_set_attributes (col, render,
                                         "pixbuf", CALL_COLUMN,
                                         "sensitive", CALL_SENSITIVE_COLUMN,
                                         NULL);
    self->priv->column_call = col;
    tapiocaui_list_append_collumn (TAPIOCAUI_LIST (self), col);

    g_signal_connect_object (G_OBJECT (render),
                             "activate",
                             G_CALLBACK (tapiocaui_buddy_list_img_call_activate_cb),
                             self,
                             0);

    tapiocaui_list_set_search_func (TAPIOCAUI_LIST (self),
                                    BUDDY_COLUMN,
                                    tapiocaui_buddy_list_search_func,
                                    NULL);

    tapiocaui_buddy_list_create_popup_menu (self);

    g_signal_connect (tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self)),
                      "button-press-event",
                      (GCallback) tapiocaui_buddy_list_button_pressed_cb,
                      self);

    g_signal_connect_object (G_OBJECT (self),
                             "iter-selected",
                             G_CALLBACK (tapiocaui_buddy_list_iter_selected_cb),
                             self,
                             0);
    g_signal_connect_object (G_OBJECT (self),
                             "iter-unselected",
                             G_CALLBACK (tapiocaui_buddy_list_iter_unselected_cb),
                             self,
                             0);
    g_signal_connect_object (G_OBJECT (self),
                             "iter-activated",
                             G_CALLBACK (tapiocaui_buddy_list_iter_activated_cb),
                             self,
                             0);

    g_signal_connect_object (G_OBJECT (self),
                             "size-allocate",
                             G_CALLBACK (tapiocaui_buddy_list_resize_cb),
                             self,
                             0);

    return G_OBJECT (self);
}

static void
tapiocaui_buddy_list_finalize (GObject *object)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (object);
    gushort i;

    g_object_unref (self->priv->img_call_active);
    for (i = 0; i <= TAPIOCA_PRESENCE_STATUS_LAST; ++i)
        g_object_unref (self->priv->img_status[i]);

    self->priv->db = NULL;
    g_free (self->priv);

    G_OBJECT_CLASS (parent_class)->finalize (object);
}

TapiocaUIBuddyList *
tapiocaui_buddy_list_new (void)
{
    TapiocaUIBuddyList *self;
    self = g_object_new (TAPIOCAUI_TYPE_BUDDY_LIST,
                         "hadjustment", NULL,
                         "vadjustment", NULL,
                         "hscrollbar_policy", GTK_POLICY_AUTOMATIC,
                         "vscrollbar_policy", GTK_POLICY_AUTOMATIC,
                         "shadow_type", GTK_SHADOW_IN,
                         NULL);
    return self;
}

void
tapiocaui_buddy_list_connect_db (TapiocaUIBuddyList *self,
                                 TapiocaUIDB *db)
{
    GList *lst_db;

    g_return_if_fail (self->priv->db == NULL);

    gtk_list_store_clear (self->priv->buddy_store);

    lst_db = tapiocaui_db_get_buddy_list (db, TAPIOCAUI_DB_ENTRY_BUDDY);
    tapiocaui_buddy_list_append_list (self, lst_db);
    self->priv->db = db;

    g_signal_connect (G_OBJECT (db),
                      "db-changed",
                      G_CALLBACK (tapiocaui_buddy_list_db_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (db),
                      "db-buddy-added",
                      G_CALLBACK (tapiocaui_buddy_list_db_buddy_added_cb),
                      self);
    g_signal_connect (G_OBJECT (db),
                      "db-buddy-removed",
                      G_CALLBACK (tapiocaui_buddy_list_db_buddy_removed_cb),
                      self);
}

void
tapiocaui_buddy_list_disconnect_db (TapiocaUIBuddyList *self)
{
    self->priv->db = NULL;
}

void
tapiocaui_buddy_list_append_buddy (TapiocaUIBuddyList *self,
                                   TapiocaUIBuddyData *buddy)
{
    TapiocaUIBuddyPresence *tmp;
    gushort status = TAPIOCA_PRESENCE_STATUS_OFFLINE;
    gboolean sensitive = FALSE;
    gchar *description;
    GtkTreeIter iter;

    tmp = g_hash_table_find (buddy->jids, (GHRFunc) gtk_true, NULL);
    if (tmp)
        status = tmp->status;

    tmp = g_hash_table_find (buddy->jids, (GHRFunc) tapiocaui_db_voice_cap_hash_filter, NULL);
    if (tmp && tmp->voice_cap)
        sensitive = TRUE;

    description = tapiocaui_buddy_list_get_description (buddy);

    tapiocaui_buddy_list_insert_sorted (self, buddy, presence_priority[status], &iter);
    gtk_list_store_set (self->priv->buddy_store,
                        &iter,
                        BUDDY_COLUMN, buddy,
                        STATUS_COLUMN, status,
                        STATUS_ICON_COLUMN, self->priv->img_status[status],
                        NAME_COLUMN, description,
                        CALL_COLUMN, self->priv->img_call_active,
                        CALL_SENSITIVE_COLUMN, sensitive,
                        -1);

    g_free (description);
}

void
tapiocaui_buddy_list_append_list (TapiocaUIBuddyList *self,
                                  GList *lst_buddy)
{
    GList *lst = lst_buddy;
    TapiocaUIBuddyData *buddy = NULL;

    while (lst) {
        buddy = (TapiocaUIBuddyData *) (lst->data);
        tapiocaui_buddy_list_append_buddy (self, buddy);
        lst = lst->next;
    }
}

void
tapiocaui_buddy_list_remove_buddy (TapiocaUIBuddyList *self,
                                   TapiocaUIBuddyData *buddy)
{
    GtkTreeIter iter;

    if (tapiocaui_buddy_list_get_buddy_iter (self, buddy, &iter))
        gtk_list_store_remove (self->priv->buddy_store, &iter);
}

TapiocaUIBuddyData *
tapiocaui_buddy_list_get_selected_buddy (TapiocaUIBuddyList *self)
{
    TapiocaUIBuddyData *buddy = NULL;
    GtkTreeModel *tree_model = GTK_TREE_MODEL (self->priv->buddy_store);
    GtkTreeIter iter;

    if (gtk_tree_selection_get_selected (
                tapiocaui_list_get_selection (TAPIOCAUI_LIST (self)),
                (GtkTreeModel **) &tree_model,
                &iter)) {
         gtk_tree_model_get (tree_model, &iter, BUDDY_COLUMN, &buddy, -1);
    }

    return buddy;
}

static gboolean
tapiocaui_buddy_list_search_func (GtkTreeModel *model,
                                  gint column,
                                  const gchar *key,
                                  GtkTreeIter *iter,
                                  gpointer search_data)
{
    TapiocaUIBuddyData *buddy;
    guint key_length, buddy_length;

    gtk_tree_model_get (model, iter, column, &buddy, -1);

    key_length = strlen (key);
    buddy_length = strlen (buddy->name);
    return (g_ascii_strncasecmp (buddy->name, key, (key_length < buddy_length ? key_length : buddy_length)) != 0);
    return TRUE;
}

static gchar *
tapiocaui_buddy_list_get_description (TapiocaUIBuddyData *buddy)
{
    TapiocaUIBuddyPresence *tmp;

    g_return_val_if_fail (buddy != NULL, NULL);

    if (g_hash_table_size (buddy->jids) == 0) {
        if ((buddy->subscription_state == TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_NONE) ||
            (buddy->subscription_state == TAPIOCA_CONTACT_LIST_SUBSCRIPTION_STATE_TO))
            return g_strdup_printf("%s", buddy->name);
        else
            return g_strdup_printf("%s", buddy->name);
    }
    else {
        tmp = g_hash_table_find (buddy->jids, (GHRFunc) gtk_true, NULL);
        if (tmp->status_desc)
            return g_markup_printf_escaped ("<b>%s</b> - <small>%s</small>", buddy->name, tmp->status_desc); 
        else
            return g_markup_printf_escaped ("<b>%s</b>", buddy->name);
    }
}

static void
tapiocaui_buddy_list_iter_selected_cb (TapiocaUIList *self,
                                       GtkTreeIter *iter,
                                       gpointer user_data)
{
    TapiocaUIBuddyData *buddy;

    gtk_tree_model_get (GTK_TREE_MODEL (TAPIOCAUI_BUDDY_LIST (self)->priv->buddy_store),
                        iter, BUDDY_COLUMN, &buddy, -1);
    g_signal_emit (G_OBJECT (self),
                   tapiocaui_buddy_list_signals [BUDDY_SELECTION_CHANGED], 0,
                   buddy, TRUE);
}

static void
tapiocaui_buddy_list_iter_unselected_cb (TapiocaUIList *self,
                                         GtkTreeIter *iter,
                                         gpointer user_data)
{
    TapiocaUIBuddyData *buddy;

    gtk_tree_model_get (GTK_TREE_MODEL (TAPIOCAUI_BUDDY_LIST (self)->priv->buddy_store),
                        iter, BUDDY_COLUMN, &buddy, -1);
    g_signal_emit (G_OBJECT (self),
                   tapiocaui_buddy_list_signals [BUDDY_SELECTION_CHANGED], 0,
                   buddy, FALSE);
}

static void
tapiocaui_buddy_list_db_buddy_added_cb (TapiocaUIDB *db,
                                        gpointer buddy,
                                        gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    tapiocaui_buddy_list_append_buddy (self, buddy);
}

static void
tapiocaui_buddy_list_db_buddy_removed_cb (TapiocaUIDB *db,
                                          gpointer buddy,
                                          gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    tapiocaui_buddy_list_remove_buddy (self, buddy);
}

static void
tapiocaui_buddy_list_db_changed_cb (TapiocaUIDB *db,
                                    gpointer user_data)
{
    gtk_widget_queue_draw (GTK_WIDGET (user_data));
}

static void
tapiocaui_buddy_list_iter_activated_cb (TapiocaUIList *self,
                                        GtkTreeIter *iter,
                                        gpointer user_data)
{
    TapiocaUIBuddyData *buddy;

    gtk_tree_model_get (GTK_TREE_MODEL (TAPIOCAUI_BUDDY_LIST (self)->priv->buddy_store),
                        iter, BUDDY_COLUMN, &buddy, -1);
    g_signal_emit (G_OBJECT (self),
                   tapiocaui_buddy_list_signals [BUDDY_ACTIVATED], 0,
                   buddy, FALSE);
}

static void
tapiocaui_buddy_list_img_call_activate_cb (TapiocaUICellRendererPixbuf *cell,
                                           const gchar *path_str,
                                           gpointer user_data)
{
    GtkTreeIter iter;
    GtkTreePath *path;
    TapiocaUIBuddyData *buddy;
    TapiocaUIBuddyPresence *tmp;
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);

    path = gtk_tree_path_new_from_string (path_str);

    gtk_tree_model_get_iter (GTK_TREE_MODEL (self->priv->buddy_store),
                             &iter, path);
    gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &iter,
                        BUDDY_COLUMN, &buddy, -1);
    tmp = g_hash_table_find (buddy->jids, (GHRFunc) tapiocaui_db_voice_cap_hash_filter, NULL);

    if ((tmp) && (tmp->voice_cap))
        g_signal_emit (G_OBJECT (self), tapiocaui_buddy_list_signals [BUDDY_ACTIVATED], 0, buddy, TRUE);

    gtk_tree_path_free  (path);
}

static void
tapiocaui_buddy_list_resize_cb (GtkWidget *widget,
                                GtkAllocation *allocation,
                                gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    GdkRectangle rect;

    gtk_tree_view_get_visible_rect (tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self)), &rect);
    if (self->priv->old_width != rect.width) {
#ifdef WITH_HILDON
        gtk_tree_view_column_set_fixed_width (self->priv->column_name, rect.width - 70);
#else
        gtk_tree_view_column_set_fixed_width (self->priv->column_name, rect.width - 26);
#endif
        self->priv->old_width = rect.width;
    }
}

static gboolean
tapiocaui_buddy_list_get_buddy_iter (TapiocaUIBuddyList *self,
                                     TapiocaUIBuddyData *buddy,
                                     GtkTreeIter *iter)
{
    TapiocaUIBuddyData *current_buddy;
    GtkTreeIter current_iter;
    gboolean valid;

    valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter);
    while (valid) {
        gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter,
                            BUDDY_COLUMN, &current_buddy, -1);
        if (current_buddy == buddy) {
            *iter = current_iter;
            return TRUE;
        }
        valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter);
    }

    return FALSE;
}
static void
tapiocaui_buddy_list_insert_sorted (TapiocaUIBuddyList *self,
                                    TapiocaUIBuddyData *buddy,
                                    guint priority,
                                    GtkTreeIter *iter)
{
    GtkTreeIter current_iter;
    const TapiocaUIBuddyData *current_buddy;
    guint current_status, current_priority;
    gint cmp = 0;

    if (!gtk_tree_model_iter_children (GTK_TREE_MODEL (self->priv->buddy_store),
                &current_iter, NULL)) {
        gtk_list_store_append (self->priv->buddy_store, iter);
        return;
    }

    do {
        gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter,
                STATUS_COLUMN, &current_status, -1);

        cmp = 0;
        current_priority = presence_priority[current_status];
        if (priority == current_priority) {
            gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter,
                                BUDDY_COLUMN, &current_buddy, -1);
            cmp = g_ascii_strncasecmp (buddy->name, current_buddy->name, strlen (buddy->name));
        }

        if (cmp < 0 || priority < current_priority) {
            gtk_list_store_insert_before (self->priv->buddy_store,
                                         iter, &current_iter);
            return;
        }
    } while (gtk_tree_model_iter_next (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter));

    gtk_list_store_append (self->priv->buddy_store, iter);
}

static void
tapiocaui_buddy_list_sort_buddy (TapiocaUIBuddyList *self,
                                 TapiocaUIBuddyData *buddy,
                                 guint priority,
                                 GtkTreeIter *iter)
{
    GtkTreeIter current_iter;
    const TapiocaUIBuddyData *current_buddy;
    guint current_status, current_priority;
    gint cmp;

    gtk_tree_model_iter_children (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter, NULL);
    do {
        gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter,
                            STATUS_COLUMN, &current_status, -1);

        cmp = 0;
        current_priority = presence_priority[current_status];
        if (priority == current_priority) {
            gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter,
                                BUDDY_COLUMN, &current_buddy, -1);
            cmp = strcasecmp (buddy->name, current_buddy->name);
        }

        if (cmp < 0 || (priority < current_priority)) {
            gtk_list_store_move_before (self->priv->buddy_store,
                                        iter, &current_iter);
            return;
        }
    } while (gtk_tree_model_iter_next (GTK_TREE_MODEL (self->priv->buddy_store), &current_iter));
}

void
tapiocaui_buddy_list_update_buddy (TapiocaUIBuddyList *self,
                                   TapiocaUIBuddyData *buddy)
{
    TapiocaUIBuddyPresence *tmp;
    gushort status = TAPIOCA_PRESENCE_STATUS_OFFLINE;
    gboolean sensitive = FALSE;
    gchar *description;
    GtkTreeIter iter;

    tmp = g_hash_table_find (buddy->jids, (GHRFunc) gtk_true, NULL);
    if (tmp)
        status = tmp->status;

    tmp = g_hash_table_find (buddy->jids, (GHRFunc) tapiocaui_db_voice_cap_hash_filter, NULL);
    if (tmp && tmp->voice_cap)
        sensitive = TRUE;

    description = tapiocaui_buddy_list_get_description (buddy);

    tapiocaui_buddy_list_get_buddy_iter (self, buddy, &iter);
    gtk_list_store_set (self->priv->buddy_store,
                        &iter,
                        STATUS_COLUMN, status,
                        STATUS_ICON_COLUMN, self->priv->img_status[status],
                        NAME_COLUMN, description,
                        CALL_SENSITIVE_COLUMN, sensitive,
                        -1);

    tapiocaui_buddy_list_sort_buddy (self, buddy, presence_priority[status], &iter);

    g_free (description);
}

static void
tapiocaui_buddy_list_create_popup_menu (TapiocaUIBuddyList *self)
{
    GtkWidget *menuitem;
    GtkWidget *image;
    GdkPixbuf *pixbuf;

    self->priv->menu = gtk_menu_new ();

    /* Chat */
    self->priv->chat_menuitem = gtk_image_menu_item_new_with_label (_("Open Chat Window"));
    g_signal_connect (self->priv->chat_menuitem, "activate",
                      (GCallback) tapiocaui_buddy_list_menu_item_open_chat_activated_cb,
                      self);
    
    pixbuf = gdk_pixbuf_new_from_file_at_scale (tapiocaui_file ("chat.png"),
                                                16, 16, TRUE, NULL);
    image = gtk_image_new_from_pixbuf (pixbuf);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (self->priv->chat_menuitem), image);
    
    gtk_menu_shell_append (GTK_MENU_SHELL (self->priv->menu), self->priv->chat_menuitem);

    /* Call */
    self->priv->call_menuitem = gtk_image_menu_item_new_with_label (_("Call"));
    g_signal_connect (self->priv->call_menuitem, "activate",
                      (GCallback) tapiocaui_buddy_list_menu_item_call_activated_cb, self);

    pixbuf = gdk_pixbuf_new_from_file_at_scale (tapiocaui_file ("call.png"),
                                                16, 16, TRUE, NULL);
    image = gtk_image_new_from_pixbuf (pixbuf);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (self->priv->call_menuitem), image);

    gtk_menu_shell_append (GTK_MENU_SHELL (self->priv->menu), self->priv->call_menuitem);

    /* Info */
    menuitem = gtk_image_menu_item_new_with_label (_("Get Info"));
    gtk_widget_set_sensitive (menuitem, FALSE);
    image = gtk_image_new_from_stock ("gtk-info", GTK_ICON_SIZE_MENU);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem), image);
    gtk_menu_shell_append (GTK_MENU_SHELL (self->priv->menu), menuitem);

    /* Separator */
    menuitem = gtk_separator_menu_item_new ();
    gtk_menu_shell_append (GTK_MENU_SHELL (self->priv->menu), menuitem);

    /* Remove */
    menuitem = gtk_image_menu_item_new_with_label (_("Remove"));
    g_signal_connect (menuitem, "activate",
                      (GCallback) tapiocaui_buddy_list_menu_item_remove_activated_cb, self);
    
    image = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_MENU);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem), image);

    gtk_menu_shell_append (GTK_MENU_SHELL (self->priv->menu), menuitem);

    gtk_widget_show_all (self->priv->menu);
}

static void
tapiocaui_buddy_list_menu_item_open_chat_activated_cb (GtkWidget *menuitem,
                                                       gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    GtkTreeView *treeview = tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self));
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;

    selection = gtk_tree_view_get_selection (treeview);
    model = gtk_tree_view_get_model (treeview);
    gtk_tree_selection_get_selected (selection,
                                     &model,
                                     &iter);

    tapiocaui_buddy_list_iter_activated_cb (TAPIOCAUI_LIST (self), &iter, user_data);
}

static void
tapiocaui_buddy_list_menu_item_call_activated_cb (GtkWidget *menuitem,
                                                  gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    GtkTreeView *treeview = tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self));
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;
    TapiocaUIBuddyData *buddy;

    selection = gtk_tree_view_get_selection (treeview);
    model = gtk_tree_view_get_model (treeview);
    gtk_tree_selection_get_selected (selection,
                                     &model,
                                     &iter);

    gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &iter,
                        BUDDY_COLUMN, &buddy, -1);

    g_signal_emit (G_OBJECT (self),
                   tapiocaui_buddy_list_signals [BUDDY_ACTIVATED],
                   0, buddy, TRUE);
}

static void
tapiocaui_buddy_list_menu_item_remove_activated_cb (GtkWidget *menuitem,
                                                    gpointer user_data)
{
    TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
    GtkTreeView *treeview = tapiocaui_list_get_treeview (TAPIOCAUI_LIST (self));
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;
    TapiocaUIBuddyData *buddy;

    selection = gtk_tree_view_get_selection (treeview);
    model = gtk_tree_view_get_model (treeview);
    gtk_tree_selection_get_selected (selection,
                                     &model,
                                     &iter);

    gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &iter,
                        BUDDY_COLUMN, &buddy, -1);

    g_signal_emit (G_OBJECT (self),
                   tapiocaui_buddy_list_signals [BUDDY_REMOVE_REQUESTED],
                   0, buddy);
}

static gboolean
tapiocaui_buddy_list_button_pressed_cb (GtkWidget *treeview,
                                        GdkEventButton *event,
                                        gpointer user_data)
{
    if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
        GtkTreeSelection *selection;

        selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
        if (gtk_tree_selection_count_selected_rows (selection) <= 1) {
            GtkTreePath *path;
            if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
                        (gint) event->x,
                        (gint) event->y,
                        &path, NULL, NULL, NULL)) {
                TapiocaUIBuddyList *self = TAPIOCAUI_BUDDY_LIST (user_data);
                TapiocaUIBuddyData *buddy;
                TapiocaUIBuddyPresence *tmp;
                GtkTreeIter iter;

                gtk_tree_selection_unselect_all (selection);
                gtk_tree_selection_select_path (selection, path);

                gtk_tree_model_get_iter (GTK_TREE_MODEL (self->priv->buddy_store),
                                         &iter, path);
                gtk_tree_path_free (path);
                gtk_tree_model_get (GTK_TREE_MODEL (self->priv->buddy_store), &iter,
                                    BUDDY_COLUMN, &buddy, -1);

                /* disable chat menu if user is offline */
                if (g_hash_table_size (buddy->jids) > 0)
                    gtk_widget_set_sensitive (self->priv->chat_menuitem, TRUE);
                else
                    gtk_widget_set_sensitive (self->priv->chat_menuitem, FALSE);

                /* disable call menu if user cannot receive calls */
                tmp = g_hash_table_find (buddy->jids, (GHRFunc) tapiocaui_db_voice_cap_hash_filter, NULL);
                if (tmp && tmp->voice_cap)
                    gtk_widget_set_sensitive (self->priv->call_menuitem, TRUE);
                else
                    gtk_widget_set_sensitive (self->priv->call_menuitem, FALSE);

                gtk_menu_popup (GTK_MENU (self->priv->menu), NULL, NULL, NULL, NULL,
                                event->button,
                                gdk_event_get_time ((GdkEvent*) event));
                return TRUE;
            }
        }
    }

    return FALSE;
}
