/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "tapiocaui_shell_presence.h"

void
tapiocaui_shell_presence_status_up_cb (TapiocaPresence *presence,
                                       const gchar *jid,
                                       guint status,
                                       const gchar *status_desc,
                                       gboolean voice_cap,
                                       gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *bare_jid;
    TapiocaUIBuddyData *buddy;
    TapiocaUIBuddyPresence *tmp;
    gchar *aux = NULL;

    bare_jid = tapiocaui_get_bare_jid (jid);

    buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                           TAPIOCAUI_DB_ENTRY_BUDDY,
                                           bare_jid);
    if (buddy) {
        if (status == TAPIOCA_PRESENCE_STATUS_OFFLINE)
            g_hash_table_remove (buddy->jids, jid);
        else {
            tmp = g_hash_table_lookup (buddy->jids, jid);
            if (!tmp) {
                tmp = g_new0 (TapiocaUIBuddyPresence, 1);
                g_hash_table_insert (buddy->jids, g_strdup (jid), tmp);
            }

            if (tmp->status_desc)
                g_free (tmp->status_desc);

            tmp->status = status;
            tmp->voice_cap = voice_cap;

            if (!status_desc)
                tmp->status_desc = NULL;
            else {
                aux = g_strdup (status_desc);
                aux = g_strstrip (aux);
                if (strlen (aux) == 0) {
                    tmp->status_desc = NULL;
                    g_free (aux);
                }
                else
                    tmp->status_desc = aux;
            }
        }

        tapiocaui_main_win_update_buddy (TAPIOCAUI_MAIN_WIN (self->priv->win),
                                         buddy);
    }
    else {
        tapiocaui_debug ("internal error");
    }

    g_free (bare_jid);
}

void
tapiocaui_shell_presence_registered_cb (TapiocaPresence *presence,
                                        gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

void
tapiocaui_shell_presence_unregistered_cb (TapiocaPresence *presence,
                                          gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}
