/*
 * Tapiocaui app
 * Copyright (C) 2005 INdT
 * @author Renato Araujo Oliveira Filho <renato.filho@indt.org.br>
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 * @author Abner Jose de Faria Silva <abner.silva@indt.org.br>
 * @author Marcio Macedo <marcio.macedo@indt.org.br>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA02111-1307USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "tapiocaui_shell_session.h"
#include "tapiocaui_shell_media.h"
#ifdef ENABLE_SYSTRAY
#include "tapiocaui_systray.h"
#endif

/* here will be make all state change process */
void
tapiocaui_shell_state_changed_cb (TapiocaConnection *tc,
                                  guint state,
                                  gpointer user_data)
{
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    guint old_state;

    switch (state) {
        case TAPIOCA_CONNECTION_STATE_DISCONNECTED:
            old_state = tapioca_connection_get_state (tc);
            if (old_state == TAPIOCA_CONNECTION_STATE_CONNECTING)
                tapiocaui_main_win_cancel_login (TAPIOCAUI_MAIN_WIN(self->priv->win), -1);
            else
                tapiocaui_main_win_logout (TAPIOCAUI_MAIN_WIN (self->priv->win));
            break;

        case TAPIOCA_CONNECTION_STATE_CONNECTED: {
            tapiocaui_shell_complete_login (self, tapioca_connection_get_id (tc));

            if (self->priv->presence) {
                guint i = 0;
                gchar **contacts = NULL;
                const gchar *status_desc = "";
                GList *buddy_list;

                tapioca_presence_subscribe (self->priv->presence, TAPIOCA_PRESENCE_STATUS_AVAILABLE, status_desc);

                buddy_list = tapiocaui_db_get_buddy_list (self->priv->db_buddy,
                                                          TAPIOCAUI_DB_ENTRY_BUDDY);
                if (buddy_list == NULL)
                    return;

                contacts = g_new0 (gchar *, g_list_length (buddy_list) + 1);
                for (; buddy_list; buddy_list = g_list_next (buddy_list))
                    contacts[i++] = g_strdup (((TapiocaUIBuddyData *) buddy_list->data)->id);
                tapioca_presence_request_status (self->priv->presence, (const gchar **) contacts);
                g_strfreev (contacts);
            }
#ifdef ENABLE_SYSTRAY            
            tapiocaui_systray_set_icon_online();
#endif
            break;
        }

        default:
            break;
    }
}

/* here will be processed all incoming session */
void
tapiocaui_shell_incoming_session_cb (TapiocaConnection *tc,
                                     const gchar *from,
                                     guint type,
                                     TapiocaSession *session,
                                     gpointer user_data)
{
    TapiocaUIShellActiveComm *comm;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);

    if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
        if (self->priv->in_use) {
            tapioca_audio_session_reject (TAPIOCA_AUDIO_SESSION (session));
            return;
        }
    }

    /* Connect session signals */
    g_signal_connect (G_OBJECT (session),
                      "member-added",
                      G_CALLBACK (tapiocaui_shell_session_member_added_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "member-removed",
                      G_CALLBACK (tapiocaui_shell_session_member_removed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "add-member-error",
                      G_CALLBACK (tapiocaui_shell_session_add_member_er_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "state-changed",
                      G_CALLBACK (tapiocaui_shell_session_state_changed_cb),
                      self);
    g_signal_connect (G_OBJECT (session),
                      "error",
                      G_CALLBACK (tapiocaui_shell_session_error_signal_cb),
                      self);

    comm = tapiocaui_shell_get_communication (self, from);

    if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
        gchar *local_cap = NULL, *remote_cap = NULL;
        TapiocaUIBuddyData *buddy = NULL;
        gchar *bare_jid = NULL;

        comm->audio_session = TAPIOCA_AUDIO_SESSION (session);

        setup_rtp_session (self, comm);
        setup_rtp_stream (self, comm);

        farsight_stream_prepare_transports (comm->stream);

        if (tapioca_audio_session_get_capabilities (comm->audio_session,
                    &local_cap,
                    &remote_cap)) {
            remote_capabilities_updated_cb (
                    comm->audio_session,
                    remote_cap,
                    comm);
        }

        /* Connect audio session signals */
        g_signal_connect (G_OBJECT (comm->audio_session),
                          "remote-capabilities-updated",
                          G_CALLBACK (remote_capabilities_updated_cb),
                          comm);
        g_signal_connect (G_OBJECT (comm->audio_session),
                          "remote-candidate-added",
                          G_CALLBACK (remote_candidate_added_cb),
                          comm);

        bare_jid = tapiocaui_get_bare_jid (from);

        buddy = tapiocaui_db_find_buddy_by_id (self->priv->db_buddy,
                                               TAPIOCAUI_DB_ENTRY_BUDDY,
                                               bare_jid);

        tapiocaui_communication_set_status (comm->comm_win, TAPIOCAUI_CALL_INCOMING);

        g_free (bare_jid);
        g_free (local_cap);
        g_free (remote_cap);
    }
    else {
        TapiocaTextSession *text_session = TAPIOCA_TEXT_SESSION (session);
        comm->text_sessions = g_list_append (comm->text_sessions, text_session);

        /* Connect text session signals */
        g_signal_connect (G_OBJECT (text_session),
                          "incoming",
                          G_CALLBACK (tapiocaui_shell_text_session_incoming_cb),
                          self);
        g_signal_connect (G_OBJECT (text_session),
                          "delivery-error",
                          G_CALLBACK (tapiocaui_shell_text_session_derror_cb),
                          self);
        tapioca_text_session_disable_message_queue (text_session);
    }
#ifdef ENABLE_SYSTRAY    
    tapiocaui_systray_set_icon_calling();
#endif    
}

/* here will be processed all connection errors */
void
tapiocaui_shell_connection_error_cb (TapiocaConnection *tc,
                                     guint code,
                                     gpointer user_data)
{

    /*guint response;*/
    GString *err_msg;

    TapiocaUIShell   *self = TAPIOCAUI_SHELL (user_data);

    switch(code)
    {
        case TAPIOCA_CONNECTION_ERROR_NOT_CONNECTED:
            err_msg = g_string_new("Not Connected");
        break;
        case TAPIOCA_CONNECTION_ERROR_NETWORK_ERROR:
            err_msg = g_string_new(_("A network error has ocurred during login. Check your network settings."));
        break;
        case TAPIOCA_CONNECTION_ERROR_AUTHENTICATION_FAILED:
            err_msg = g_string_new(_("User name and password do not match. Verify your login information."));
        break;
        default:
            err_msg = g_string_new(_("An unkown error has ocurred."));
        break;
    }

    /*TODO: disconnect this signal on a better apropriate funcion */
    g_signal_handlers_disconnect_by_func (G_OBJECT (self->priv->tc),
                                          G_CALLBACK (tapiocaui_shell_connection_error_cb),
                                          self);

    tapiocaui_main_win_set_error_message(TAPIOCAUI_MAIN_WIN (self->priv->win), err_msg->str);
    g_string_free(err_msg, TRUE);
}

void
tapiocaui_shell_session_member_added_cb (TapiocaSession *session,
                                         const gchar *uri,
                                         gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

void
tapiocaui_shell_session_member_removed_cb (TapiocaSession *session,
                                           const gchar *uri,
                                           gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

void
tapiocaui_shell_session_add_member_er_cb (TapiocaSession *session,
                                          guint code,
                                          gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

void
tapiocaui_shell_session_state_changed_cb (TapiocaSession *session,
                                          guint state,
                                          gpointer user_data)
{
    gchar **session_members = NULL;
    TapiocaUIShellActiveComm *active_comm = NULL;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    guint type;

    session_members = tapioca_session_get_members (session);

    if (!session_members) {
        tapiocaui_debug ("No member in session");
        return;
    }

    active_comm = tapiocaui_shell_get_communication (self, session_members[0]);

    type = tapioca_session_get_session_type (session);

    switch (state) {
        case TAPIOCA_SESSION_STATE_INITIATING:
            if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
                farsight_stream_prepare_transports (active_comm->stream);
            }
            break;
        case TAPIOCA_SESSION_STATE_ESTABLISHED:
            if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
                tapiocaui_shell_open_call (self, active_comm);
            }
            break;
        case TAPIOCA_SESSION_STATE_CLOSED:
            if (type == TAPIOCA_SESSION_TYPE_AUDIO) {
                tapiocaui_shell_close_call (self, active_comm);
            }
            else if (type == TAPIOCA_SESSION_TYPE_TEXT) {
                gchar *msg = g_strdup_printf (_("%s has closed the conversation window"), active_comm->buddy->name);
                tapiocaui_communication_append_information_message (active_comm->comm_win,
                        0, msg, FALSE);
                g_free (msg);
                active_comm->text_sessions = g_list_remove (active_comm->text_sessions, session);
            }
            g_object_unref (session);
            break;

        default:
            break;
    }

    g_strfreev (session_members);
}

void
tapiocaui_shell_session_error_signal_cb (TapiocaSession *session,
                                         guint code,
                                         gpointer user_data)
{
    tapiocaui_debug ("Not implemented");
}

void
tapiocaui_shell_text_session_incoming_cb (TapiocaTextSession *session,
                                          const gchar *from,
                                          guint type,
                                          const gchar *subject,
                                          const gchar *body,
                                          guint timestamp,
                                          TapiocaTextSessionNotificationState notification_state,
                                          gpointer user_data)
{
    TapiocaUIShellActiveComm *comm = NULL;
    TapiocaUIShell *self = TAPIOCAUI_SHELL (user_data);
    gchar *msg = NULL;
    gboolean replaceable = TRUE;

    comm = tapiocaui_shell_get_communication (self, from);
    if (type == TAPIOCA_TEXT_SESSION_MESSAGE_TYPE_PLAIN_TEXT) {
        if (comm) {
            comm->current_text_session = session;
            tapiocaui_communication_append_message (comm->comm_win, timestamp,
                    comm->buddy->name, body);
        }
    }

    switch (notification_state) {
        case TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_ACTIVE:
            tapiocaui_communication_remove_replaceable_notification (comm->comm_win);
            break;
        case TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_COMPOSING:
            msg = g_strdup_printf (_("%s is typing ..."), comm->buddy->name);
            break;
        case TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_PAUSED:
            tapiocaui_communication_remove_replaceable_notification (comm->comm_win);
            break;
        case TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_INACTIVE:
            msg = g_strdup_printf (_("%s is idle"), comm->buddy->name);
            break;
        case TAPIOCA_TEXT_SESSION_NOTIFICATION_STATE_GONE:
            msg = g_strdup_printf (_("%s has closed the conversation window"), comm->buddy->name);
            replaceable = FALSE;
            break;
        default:
            break;
    }

    if (msg) {
        tapiocaui_communication_append_information_message (comm->comm_win,
                timestamp, msg, replaceable);
        g_free (msg);
    }
}

void
tapiocaui_shell_text_session_derror_cb (TapiocaTextSession *session,
                                        const gchar        *from,
                                        guint               type,
                                        const gchar        *subject,
                                        const gchar        *body,
                                        guint               timestamp,
                                        guint               code,
                                        gpointer            user_data)
{
    tapiocaui_debug ("Not implemented");
}
